/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "Reconciliation.h"
using pecunia::currency::Money;
using ::QDate;
using std::map;
using drn::banking::BankAccount;
using drn::banking::ReconciledBankAccount;
using drn::foundation::Optional;

#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <QObject>
using ::QObject;

#include <algorithm>
using std::find_if;
#include <ostream>
using std::ostream;
#include <tuple>
using std::tie;
#include <utility>
using std::move;

#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include "BankingErrors.h"
using drn::banking::BankError;
#include <foundation/PresentationText.h>
using drn::foundation::presentationText;


ReconciledBankAccount::ReconciledBankAccount(
	BankAccount bankAccount,
	const Iso4217Codes& currency
) :
	bankAccount_{move(bankAccount)},
	balance_{currency},
	reconciledOn_{}
{}

ReconciledBankAccount::ReconciledBankAccount(
	BankAccount bankAccount,
	Money balance,
	Optional<QDate> reconciledOn
) :
	bankAccount_{move(bankAccount)},
	balance_{move(balance)},
	reconciledOn_{move(reconciledOn)}
{}

const BankAccount& ReconciledBankAccount::bankAccount() const noexcept
{
	return this->bankAccount_;
}

const Money& ReconciledBankAccount::balance() const noexcept
{
	return this->balance_;
}

const Optional<QDate>& ReconciledBankAccount::reconciledOn() const noexcept
{
	return this->reconciledOn_;
}

void ReconciledBankAccount::reconcile(const QDate& reconciledOn, const Money& balance)
{
	if (reconciledOn < this->reconciledOn_)
		throw BankError{
			this->bankAccount_.bank_,
			QObject::tr(
				"The last reconciled date, %1, cannot be updated to be in the past, %2. "
			)
				.arg(presentationText(this->reconciledOn_))
				.arg(presentationText(reconciledOn))
		};
	this->reconciledOn_ = reconciledOn;
	this->balance_ = balance;
}

bool drn::banking::operator==(const ReconciledBankAccount& lhs, const ReconciledBankAccount& rhs)
{
	return tie(lhs.bankAccount(), lhs.balance(), lhs.reconciledOn())
		== tie(rhs.bankAccount(), rhs.balance(), rhs.reconciledOn());
}

bool drn::banking::operator!=(const ReconciledBankAccount& lhs, const ReconciledBankAccount& rhs)
{
	return ! (lhs == rhs);
}

ostream& drn::banking::operator<<(ostream& o, const ReconciledBankAccount& rba)
{
	o << "Bank Account: (" << rba.bankAccount() << "), Balance: (" << rba.balance()
			<< ") Reconciled On: (" << rba.reconciledOn();
	return o << ')';
}
