/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "XmlBudgetFile.h"
using std::make_tuple;
using std::move;
using std::tuple;

#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
using pecunia::currency::toIso4217Code;
using pecunia::currency::toStdString;
#include <pecunia/Money.h>
using pecunia::currency::Money;
#include <QDebug>
// using qDebug;
#include <QIODevice>
using ::QIODevice;
#include <QObject>
using ::QObject;
#include <QString>
using ::QString;
#include <QXmlStreamReader>
using ::QXmlStreamReader;
#include <QXmlStreamWriter>
using ::QXmlStreamWriter;

#include <algorithm>
using std::find_if;
#include <set>
using std::set;
#include <stdexcept>
using std::exception;
#include <map>
using std::map;
#include <vector>
using std::vector;

#include <accounting/Account.h>
using drn::accounting::Account;
using drn::accounting::toAccountTypes;
#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
using drn::accounting::AccountNumber;
#include <accounting/LedgerBooks.h>
using drn::accounting::LedgerAccount;
using drn::accounting::GeneralLedger;
#include <accounting/Transaction.h>
using drn::accounting::creditAbbreviation;
using drn::accounting::debitAbbreviation;
using drn::accounting::presentationText;
using drn::accounting::toTransactionStatuses;
using drn::accounting::Transaction;
using drn::accounting::TransactionNumber;
#include <banking/Bank.h>
using drn::banking::Bank;
#include <banking/BankAccount.h>
using drn::banking::BankAccount;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
using drn::banking::SupportedAccountTypes;
using drn::banking::presentationText;
using drn::banking::toSupportedAccountTypes;
#include <banking/BankMap.h>
using drn::banking::BankMap;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/Reconciliation.h>
using drn::banking::ReconciledBankAccount;
#include "BudgetElements.h"
using drn::file_storage::internal::AccountsElement;
using drn::file_storage::internal::ApplicationElement;
using drn::file_storage::internal::BasicElement;
using drn::file_storage::internal::BankElement;
using drn::file_storage::internal::BankAccountElement;
using drn::file_storage::internal::BanksElement;
using drn::file_storage::internal::BillElement;
using drn::file_storage::internal::BillsElement;
using drn::file_storage::internal::DebtElement;
using drn::file_storage::internal::DebtsElement;
using drn::file_storage::internal::DiscretionariesElement;
using drn::file_storage::internal::DiscretionaryElement;
using drn::file_storage::internal::GoalElement;
using drn::file_storage::internal::GoalsElement;
using drn::file_storage::internal::IndexedElement;
using drn::file_storage::internal::LedgerElement;
using drn::file_storage::internal::LedgersElement;
using drn::file_storage::internal::ReconciliationElement;
using drn::file_storage::internal::ReconciliationsElement;
using drn::file_storage::internal::WageElement;
using drn::file_storage::internal::WagesElement;
#include <budgeting/Budget.h>
using drn::budgeting::Budget;
#include <budgeting/BudgetItemTypes.h>
using drn::budgeting::BudgetItemIdentifier;
using drn::budgeting::BudgetItemTypes;
using drn::budgeting::toBudgetItemTypes;
using drn::budgeting::toWeak;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <budgeting/Nontrack.h>
using drn::budgeting::Nontrack;
#include <budgeting/EventFrequency.h>
using drn::budgeting::presentationText;
using drn::budgeting::toEventFrequency;
#include <budgeting/Bill.h>
using drn::budgeting::Bill;
#include <budgeting/Goal.h>
using drn::budgeting::Goal;
#include <budgeting/Wage.h>
using drn::budgeting::Wage;
#include <budgeting/Debt.h>
using drn::budgeting::Debt;
using drn::budgeting::Percentage;
#include <file-storage/FileStorageErrors.h>
using drn::file_storage::BudgetFileError;
#include <foundation/Error.h>
using drn::foundation::Error;
#include <foundation/ObserverPtr.hpp>
using drn::foundation::ObserverPtr;
#include <foundation/Optional.hpp>
using drn::foundation::inPlace;
using drn::foundation::Optional;
#include <foundation/TypeIndexMap.hpp>
using drn::foundation::TypeIndexMap;
#include <surveying/ReconciledSurvey.h>
using drn::surveying::ReconciledSurvey;
using drn::surveying::ReconciledSurveys;


namespace
{

class BudgetItemIdentifierIndexer
{
	std::map<BudgetItemTypes, map<IndexedElement::IdType, BudgetSource>> budgetIds_;

public:
	void addMapping(
		const BudgetItemTypes& type,
		const IndexedElement::IdType& id,
		const BudgetSource& source
	)
	{
		this->budgetIds_[type][id] = source;
	}

	const BudgetSource& lookUp(const BudgetItemTypes& type, const IndexedElement::IdType& id)
	try
	{
		return this->budgetIds_.at(type).at(id);
	}
	catch (const exception& error)
	{
		throw BudgetFileError{
			{},
			{},
			{},
			QObject::tr(
				"The budget file contains a malformed budget item ID mapping. There is not an "
					"identifier of %1 of the type %2."
			)
				.arg(id)
				.arg(presentationText(type)),
			error
		};
	}

	const IndexedElement::IdType& lookUpId(const BudgetItemTypes& type, const BudgetSource& source)
	try
	{
		const auto& types{this->budgetIds_.at(type)};
		const auto idSource{
			find_if(
				types.cbegin(),
				types.cend(),
				[&source] (const auto& IdSourceFromType)
				{
					return IdSourceFromType.second == source;
				}
			)
		};

		if (idSource == types.cend())
			throw Error{QObject::tr("Failed to find budget source.")};
		return idSource->first;
	}
	catch (const exception& error)
	{
		throw BudgetFileError{
			{},
			{},
			{},
			QObject::tr(
				"The budget file contains a malformed budget item ID mapping. There is not a "
					"budget source '%1' of the type %2."
			)
				.arg(presentationText(source))
				.arg(presentationText(type)),
			error
		};
	}
};

}

tuple<
	GeneralLedger,
	Budget,
	TypeIndexMap<BudgetSource, AccountCode>,
	BankMap,
	map<BankName, map<AccountNumber, SupportedAccountTypes>>,
	ReconciledSurveys
> drn::file_storage::internal::read(QIODevice& file)
{
	qDebug() << "Reading from budget file.";
	Q_ASSERT(file.isOpen());
	WagesElement wagesElement{};
	BillsElement billsElements{};
	DebtsElement debtsElements{};
	GoalsElement goalsElements{};
	DiscretionariesElement nontracksElement{};
	AccountsElement accountsElement{};
	LedgersElement ledgersElement{};
	BanksElement banksElement{};
	using RevisionReaders = map<QString, ObserverPtr<BasicElement>>;
	const RevisionReaders revision1Readers{
		{wagesElement.tag(), &wagesElement},
		{billsElements.tag(), &billsElements},
		{debtsElements.tag(), &debtsElements},
		{goalsElements.tag(), &goalsElements},
		{nontracksElement.tag(), &nontracksElement},
		{accountsElement.tag(), &accountsElement},
		{ledgersElement.tag(), &ledgersElement},
		{banksElement.tag(), &banksElement}
	};
	ReconciliationsElement reconciliationsElement{};
	RevisionReaders revision2Readers{
		{wagesElement.tag(), &wagesElement},
		{billsElements.tag(), &billsElements},
		{debtsElements.tag(), &debtsElements},
		{goalsElements.tag(), &goalsElements},
		{nontracksElement.tag(), &nontracksElement},
		{accountsElement.tag(), &accountsElement},
		{ledgersElement.tag(), &ledgersElement},
		{banksElement.tag(), &banksElement},
		{reconciliationsElement.tag(), &reconciliationsElement}
	};
	const map<quint8, ObserverPtr<const RevisionReaders>> readers{
		{
			{1u, &revision1Readers},
			{2u, &revision2Readers}
		}
	};
	assert(
		ApplicationElement::currentRevision_ == readers.size()
		&& "The number of budget file readers should be the same as the current revision."
	);
	ApplicationElement applicationElement{};
	const auto fileRevision{
		[&file, &applicationElement] ()
		{
			QXmlStreamReader xml{&file};
			while (xml.readNextStartElement())
			{
				if (applicationElement.tag() == xml.name().toString())
				{
					applicationElement.read(xml);
					return applicationElement.revision_;
				}
			}
			throw BudgetFileError{
				xml.errorString(),
				xml.lineNumber(),
				xml.columnNumber(),
				QObject::tr("Missing the application element in the budget file.")
			};
		}()
	};
	qDebug() << "Parsing budget file revision" << applicationElement.revision_;

	if ( ! file.reset())
		throw BudgetFileError{
			file.errorString(),
			{},
			{},
			QObject::tr("Failed to reset budget file, after reading the application element.")
		};
	QXmlStreamReader xml{&file};

	while (xml.readNextStartElement())
	{
		const auto elementName{xml.name().toString()};

		if (applicationElement.tag() == elementName)
		{
			applicationElement.read(xml);
			assert(
				applicationElement.revision_ == fileRevision && "The revision should be the same."
			);
		}
		else if (readers.at(fileRevision)->count(elementName) > 0)
			readers.at(fileRevision)->at(elementName)->read(xml);
		else
			throw BudgetFileError{
				xml.errorString(),
				xml.lineNumber(),
				xml.columnNumber(),
				QObject::tr("An unexpected error top-level element, %1, appeared in the file.")
					.arg(elementName)
			};
	}

	if (xml.hasError())
		throw BudgetFileError{
			xml.errorString(),
			xml.lineNumber(),
			xml.columnNumber(),
			QObject::tr("An unexpected error appeared in the file.")
		};
	BudgetItemIdentifierIndexer budgetIndexer{};
	Budget b{};

	for (const auto& wageElement : wagesElement.wages_)
		try
		{
			const BudgetSource source{wageElement.sourceName_};
			b.wages_.add(
				Wage{
					source,
					Money{
						wageElement.major_,
						wageElement.minor_,
						toIso4217Code(wageElement.currency_.toStdString())
					},
					toEventFrequency(wageElement.period_),
					wageElement.nextOccurOn_
				}
			);
			budgetIndexer.addMapping(BudgetItemTypes::Wage, wageElement.id_, source);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed wage data."),
				error
			};
		}

	for (const auto& billElement : billsElements.bills_)
		try
		{
			const BudgetSource source{billElement.sourceName_};
			b.bills_.add(
				Bill{
					source,
					Money{
						billElement.major_,
						billElement.minor_,
						toIso4217Code(billElement.currency_.toStdString())
					},
					toEventFrequency(billElement.period_),
					billElement.nextOccurOn_
				}
			);
			budgetIndexer.addMapping(BudgetItemTypes::Bill, billElement.id_, source);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed bill data."),
				error
			};
		}

	for (const auto& debtElement : debtsElements.debts_)
		try
		{
			const BudgetSource source{debtElement.sourceName_};
			b.debts_.add(
				Debt{
					source,
					Money{
						debtElement.minimumMajor_,
						debtElement.minimumMinor_,
						toIso4217Code(debtElement.currency_.toStdString())
					},
					toEventFrequency(debtElement.period_),
					debtElement.nextOccurOn_,
					Money{
						debtElement.totalMajor_,
						debtElement.totalMinor_,
						toIso4217Code(debtElement.currency_.toStdString())
					},
					Percentage{debtElement.interest_}
				}
			);
			budgetIndexer.addMapping(BudgetItemTypes::Debt, debtElement.id_, source);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed debt data."),
				error
			};
		}

	for (const auto& goalElement : goalsElements.goals_)
		try
		{
			const BudgetSource source{goalElement.sourceName_};
			b.goals_.add(
				Goal{
					source,
					Money{
						goalElement.major_,
						goalElement.minor_,
						toIso4217Code(goalElement.currency_.toStdString())
					},
					toEventFrequency(goalElement.period_),
					goalElement.nextOccurOn_,
					Money{
						goalElement.finalMajor_,
						goalElement.finalMinor_,
						toIso4217Code(goalElement.currency_.toStdString())
					}
				}
			);
			budgetIndexer.addMapping(BudgetItemTypes::Goal, goalElement.id_, source);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed goal data."),
				error
			};
		}

	for (const auto& nontrackElement : nontracksElement.discretionaries_)
		try
		{
			const BudgetSource source{nontrackElement.sourceName_};
			b.nontracks_.add(
				Nontrack{
					source,
					Money{
						nontrackElement.major_,
						nontrackElement.minor_,
						toIso4217Code(nontrackElement.currency_.toStdString())
					},
					toEventFrequency(nontrackElement.period_),
					nontrackElement.nextOccurOn_
				}
			);
			budgetIndexer.addMapping(BudgetItemTypes::Nontrack, nontrackElement.id_, source);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed nontrack data."),
				error
			};
		}
	map<AccountCode, LedgerAccount> ledgers{};
	// Account ID -> txnNum -> transaction
	map<IndexedElement::IdType, map<TransactionNumber, Transaction>> ledgersTransactions{};
	const auto findAccountName{
		[] (const std::vector<AccountElement>& accounts, const IndexedElement::IdType& id)
		{
			const auto account{
				find_if(
					accounts.cbegin(),
					accounts.cend(),
					[&id] (const auto& element)
					{
						return element.id_ == id;
					}
				)
			};

			if (account == accounts.cend())
				throw BudgetFileError{
					{},
					{},
					{},
					QObject::tr(
						"The budget file contains malformed account/transaction/reconciliation "
							"data. Failed to locate account name for account number %1."
					).arg(id)
				};
			return account->name_;
		}
	};

	for (const auto& ledger : ledgersElement.ledgers_)
		for (const auto& transaction : ledger.transactions_)
		{
			TransactionNumber txnNum{transaction.id_};
			const auto debitingNumber{
				debitAbbreviation() == transaction.entry_
					? transaction.accountId_
					: ledger.id_
			};
			const AccountCode debiting{
				AccountNumber{debitingNumber},
				findAccountName(accountsElement.accounts_, debitingNumber)
			};
			const auto creditingNumber{
				creditAbbreviation() == transaction.entry_
					? transaction.accountId_
					: ledger.id_,
			};
			const AccountCode crediting{
				AccountNumber{creditingNumber},
				findAccountName(accountsElement.accounts_, creditingNumber)
			};
			const Money amount{
				transaction.major_,
				transaction.minor_,
				toIso4217Code(transaction.currency_.toStdString())
			};
			Transaction txn{
				transaction.date_,
				debiting,
				crediting,
				amount,
				transaction.description_.hasValue()
					? transaction.description_
					: Optional<QString>{},
				txnNum,
				toTransactionStatuses(transaction.status_)
			};
			ledgersTransactions[ledger.id_].emplace(move(txnNum), move(txn));
		}

	for (const auto& accountElement : accountsElement.accounts_)
		try
		{
			ledgers.emplace(
				AccountCode{
					AccountNumber{accountElement.id_},
					accountElement.name_,
					accountElement.parent_.hasValue()
						? Optional<AccountNumber>{AccountNumber{*accountElement.parent_}}
						: Optional<AccountNumber>{}
				},
				LedgerAccount{
					Account{
						toAccountTypes(accountElement.type_),
						AccountCode{
							AccountNumber{accountElement.id_},
							accountElement.name_,
							accountElement.parent_.hasValue()
								? Optional<AccountNumber>{AccountNumber{*accountElement.parent_}}
								: Optional<AccountNumber>{}
						},
						Money{
							accountElement.major_,
							accountElement.minor_,
							toIso4217Code(accountElement.currency_.toStdString())
						},
						accountElement.isClosed_
					},
					ledgersTransactions[accountElement.id_]
				}
			);
		}
		catch (const exception& error)
		{
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr("The budget file contains malformed ledger data."),
				error
			};
		}
	auto openingAccount{
		[&accountsElement, &ledgersElement, &findAccountName] () -> Optional<AccountCode>
		{
			if ( ! ledgersElement.openingAccountId_.hasValue())
				return {};
			return AccountCode{
				AccountNumber{*ledgersElement.openingAccountId_},
				findAccountName(accountsElement.accounts_, *ledgersElement.openingAccountId_)
			};
		}()
	};
	map<BankName, map<AccountNumber, SupportedAccountTypes>> bankAccountTypes;
	BankMap banks{
		[&banksElement, &bankAccountTypes] ()
		{
			map<BankName, Bank> rawBanks;

			for (const auto& bankElement : banksElement.banks_)
				try
				{
					const BankName name{bankElement.name_};
					set<AccountNumber> numbers;

					for (const auto& account : bankElement.accounts_)
					{
						AccountNumber number{account.accountId_};
						numbers.emplace(number);
						bankAccountTypes[name][move(number)] = toSupportedAccountTypes(
							account.type_
						);
					}
					rawBanks[name] = Bank{name, move(numbers), bankElement.isClosed_};
				}
				catch (const exception& error)
				{
					throw BudgetFileError{
						{},
						{},
						{},
						QObject::tr("The budget file contains malformed banks data."),
						error
					};
				}
			return rawBanks;
		}()
	};
	TypeIndexMap<BudgetSource, AccountCode> budgetCodes{};

	for (const auto& wageElement : wagesElement.wages_)
		budgetCodes.add<Wage>(
			BudgetSource{wageElement.sourceName_},
			ledgers.at(AccountCode{AccountNumber{wageElement.accountId_}}).account_.code()
		);

	for (const auto& billElement : billsElements.bills_)
		budgetCodes.add<Bill>(
			BudgetSource{billElement.sourceName_},
			ledgers.at(AccountCode{AccountNumber{billElement.accountId_}}).account_.code()
		);

	for (const auto& debtElement : debtsElements.debts_)
		budgetCodes.add<Debt>(
			BudgetSource{debtElement.sourceName_},
			ledgers.at(AccountCode{AccountNumber{debtElement.accountId_}}).account_.code()
		);

	for (const auto& goalElement : goalsElements.goals_)
		budgetCodes.add<Goal>(
			BudgetSource{goalElement.sourceName_},
			ledgers.at(AccountCode{AccountNumber{goalElement.accountId_}}).account_.code()
		);

	for (const auto& nontrackElement : nontracksElement.discretionaries_)
		budgetCodes.add<Nontrack>(
			BudgetSource{nontrackElement.sourceName_},
			ledgers.at(AccountCode{AccountNumber{nontrackElement.accountId_}}).account_.code()
		);
	map<BankAccount, ReconciledSurvey> surveys{};

	for (const auto& reconciliationElement : reconciliationsElement.reconciliations_)
	{
		const AccountNumber accountNumber{reconciliationElement.accountId_};
		const auto bankInfo{
			find_if(
				bankAccountTypes.cbegin(),
				bankAccountTypes.cend(),
				[&accountNumber] (const auto& nameNumberTypes)
				{
					for (const auto& numberType : nameNumberTypes.second)
						if (numberType.first == accountNumber)
							return true;
					return false;
				}
			)
		};

		if (bankInfo == bankAccountTypes.cend())
			throw BudgetFileError{
				{},
				{},
				{},
				QObject::tr(
					"The budget file contains malformed reconciliation data, failed to locate the"
						"bank information for %1."
				).arg(presentationText(accountNumber))
			};
		const BankAccount ba{
			bankInfo->first,
			BankAccountType{
				AccountCode{
					accountNumber,
					findAccountName(accountsElement.accounts_, reconciliationElement.accountId_)
				},
				bankInfo->second.at(accountNumber)
			}
		};
		map<BudgetItemIdentifier, Money> distributions{};

		for (const auto& distribution : reconciliationElement.distributions_)
		{
			const auto type{toBudgetItemTypes(distribution.type_)};
			distributions.emplace(
				BudgetItemIdentifier{type, budgetIndexer.lookUp(type, distribution.budgetItemId_)},
				Money{
					distribution.major_,
					distribution.minor_,
					toIso4217Code(distribution.currency_.toStdString())
				}
			);
		}
		surveys.emplace(
			ba,
			ReconciledSurvey{
				ReconciledBankAccount{
					ba,
					Money{
						reconciliationElement.major_,
						reconciliationElement.minor_,
						toIso4217Code(reconciliationElement.currency_.toStdString())
					},
					reconciliationElement.date_
				},
				move(distributions),
 				reconciliationElement.checksum_
			}
		);
	}
	return make_tuple(
		GeneralLedger{move(ledgers), move(openingAccount)},
		move(b),
		move(budgetCodes),
		move(banks),
		move(bankAccountTypes),
		ReconciledSurveys{move(surveys)}
	);
}

void drn::file_storage::internal::write(
	const Budget& b,
	const GeneralLedger& gl,
	const BankMap& banks,
	const map<BankName, map<AccountNumber, SupportedAccountTypes>>& bankAccountTypes,
	const TypeIndexMap<BudgetSource, AccountCode>& budgetCodes,
	const ReconciledSurveys& surveys,
	QIODevice& file,
	const bool shouldIndent
)
{
	qDebug() << "Preparing to write to the budget file.";
	Q_ASSERT(file.isOpen());
	QXmlStreamWriter xml{&file};

	if (shouldIndent)
	{
		xml.setAutoFormatting(true);
		xml.setAutoFormattingIndent(-1); // One tab.
	}
	xml.writeStartDocument();

	qDebug() << "Filling application element";
	const ApplicationElement applicationElement{};
	BudgetItemIdentifierIndexer budgetIndexer{};
	qDebug() << "Filling wage elements";
	const WagesElement wageElements{
		[&b, &budgetCodes, &budgetIndexer] ()
		{
			WagesElement we;
			auto id{static_cast<IndexedElement::IdType>(IndexedElement::invalidId_ + 1)};

			for (const auto& wage : b.wages_)
			{
				we.wages_.emplace_back(
					id,
					wage.second.amount().major(),
					wage.second.amount().minor(),
					presentationText(wage.second.period()),
					wage.second.source().name(),
					QString::fromStdString(toStdString(wage.second.amount().code())),
					wage.second.nextOccurOn(),
					budgetCodes.value<Wage>(wage.second.source()).number().integer()
				);
				budgetIndexer.addMapping(BudgetItemTypes::Wage, id, wage.first);
				++id;
			}
			return we;
		}()
	};
	qDebug() << "Filling bill elements";
	const BillsElement billElements{
		[&b, &budgetCodes, &budgetIndexer] ()
		{
			BillsElement ee;
			auto id{static_cast<IndexedElement::IdType>(IndexedElement::invalidId_ + 1)};

			for (const auto& bill : b.bills_)
			{
				ee.bills_.emplace_back(
					id,
					bill.second.amount().major(),
					bill.second.amount().minor(),
					presentationText(bill.second.period()),
					bill.second.source().name(),
					QString::fromStdString(toStdString(bill.second.amount().code())),
					bill.second.nextOccurOn(),
					budgetCodes.value<Bill>(bill.second.source()).number().integer()
				);
				budgetIndexer.addMapping(BudgetItemTypes::Bill, id, bill.first);
				++id;
			}
			return ee;
		}()
	};
	qDebug() << "Filling debt elements";
	const DebtsElement debtsElements{
		[&b, &budgetCodes, &budgetIndexer] ()
		{
			DebtsElement le;
			auto id{static_cast<IndexedElement::IdType>(IndexedElement::invalidId_ + 1)};

			for (const auto& debt : b.debts_)
			{
				le.debts_.emplace_back(
					id,
					debt.second.borrowed().major(),
					debt.second.borrowed().minor(),
					presentationText(debt.second.period()),
					debt.second.source().name(),
					QString::fromStdString(toStdString(debt.second.amount().code())),
					debt.second.amount().major(),
					debt.second.amount().minor(),
					debt.second.interest().rate_,
					debt.second.nextOccurOn(),
					budgetCodes.value<Debt>(debt.second.source()).number().integer()
				);
				budgetIndexer.addMapping(BudgetItemTypes::Debt, id, debt.first);
				++id;
			}
			return le;
		}()
	};
	qDebug() << "Filling goal elements";
	const GoalsElement goalElements{
		[&b, &budgetCodes, &budgetIndexer] ()
		{
			GoalsElement ge;
			auto id{static_cast<IndexedElement::IdType>(IndexedElement::invalidId_ + 1)};

			for (const auto& goal : b.goals_)
			{
				ge.goals_.emplace_back(
					id,
					goal.second.amount().major(),
					goal.second.amount().minor(),
					presentationText(goal.second.period()),
					goal.second.source().name(),
					QString::fromStdString(toStdString(goal.second.amount().code())),
					goal.second.final().major(),
					goal.second.final().minor(),
					goal.second.nextOccurOn(),
					budgetCodes.value<Goal>(goal.second.source()).number().integer()
				);
				budgetIndexer.addMapping(BudgetItemTypes::Goal, id, goal.first);
				++id;
			}
			return ge;
		}()
	};
	qDebug() << "Filling discretionary elements";
	const DiscretionariesElement nontrackElements{
		[&b, &budgetCodes, &budgetIndexer] ()
		{
			DiscretionariesElement des;
			auto id{static_cast<IndexedElement::IdType>(IndexedElement::invalidId_ + 1)};

			for (const auto& nontrack : b.nontracks_)
			{
				des.discretionaries_.emplace_back(
					id,
					nontrack.second.amount().major(),
					nontrack.second.amount().minor(),
					presentationText(nontrack.second.period()),
					nontrack.second.source().name(),
					QString::fromStdString(toStdString(nontrack.second.amount().code())),
					nontrack.second.nextOccurOn(),
					budgetCodes.value<Nontrack>(nontrack.second.source()).number().integer()
				);
				budgetIndexer.addMapping(BudgetItemTypes::Nontrack, id, nontrack.first);
				++id;
			}
			return des;
		}()
	};
	AccountsElement accountsElement{};
	LedgersElement ledgersElement{
		gl.openingAccountCode().hasValue()
			? gl.openingAccountCode()->number().integer()
			: Optional<IndexedElement::IdType>{}
	};
	qDebug() << "Filling account, ledger, and transaction elements";

	for (const auto& ledger : gl.ledgers())
	{
		const auto& account{ledger.second.account_};
		accountsElement.accounts_.emplace_back(
			ledger.second.account_.code().number().integer(),
			ledger.second.account_.code().name(),
			presentationText(ledger.second.account_.type()),
			ledger.second.account_.balance().major(),
			ledger.second.account_.balance().minor(),
			QString::fromStdString(toStdString(ledger.second.account_.balance().code())),
			ledger.second.account_.isClosed(),
			account.code().parent().hasValue()
				? Optional<quint32>(account.code().parent()->integer())
				: Optional<quint32>{}
		);
		LedgerElement ledgerElement{ledger.second.account_.code().number().integer()};

		for (const auto& transaction : ledger.second.transactions_)
			ledgerElement.transactions_.emplace_back(
				transaction.second.number()->integer(),
				QDate{
					static_cast<qint32>(transaction.second.occurredOn().year()),
					static_cast<qint32>(static_cast<quint32>(
						transaction.second.occurredOn().month()
					)),
					static_cast<qint32>(static_cast<quint32>(
						transaction.second.occurredOn().day()
					))
				},
				transaction.second.crediting() == ledger.first
					? transaction.second.debiting().number().integer()
					: transaction.second.crediting().number().integer(),
				transaction.second.crediting() == ledger.first
					? debitAbbreviation()
					: creditAbbreviation(),
				transaction.second.amount().major(),
				transaction.second.amount().minor(),
				QString::fromStdString(toStdString(transaction.second.amount().code())),
				presentationText(transaction.second.status()),
				transaction.second.description().hasValue()
					? *transaction.second.description()
					: Optional<QString>{}
			);
		ledgersElement.ledgers_.emplace_back(ledgerElement);
	}
	qDebug() << "Filling bank elements";
	const BanksElement banksElement{
		[&banks, &bankAccountTypes] ()
		{
			vector<BankElement> rawBanks;

			for (const auto& bank : banks)
			{
				vector<BankAccountElement> accounts;

				for (const auto& number : bank.second.accounts())
					accounts.emplace_back(
						number.integer(),
						presentationText(bankAccountTypes.at(bank.first).at(number))
					);
				rawBanks.emplace_back(
					bank.first.institution(),
					bank.second.isClosed(),
					move(accounts)
				);
			}
			return rawBanks;
		}()
	};
	qDebug() << "Filling reconciled bank accounts elements";
	const ReconciliationsElement reconciliationsElement{
		[&surveys, &budgetIndexer] ()
		{
			ReconciliationsElement elements{};

			for (const auto& bankAccountReconciledSurvey : surveys)
			{
				const auto& bankAccount{bankAccountReconciledSurvey.first};
				const auto& reconciledSurvey{bankAccountReconciledSurvey.second};
				vector<DistributionElement> distributions{};

				for (const auto& unused : reconciledSurvey.distribution())
					distributions.emplace_back(
						toWeak(unused.first.type_),
						budgetIndexer.lookUpId(unused.first.type_, unused.first.source_),
						unused.second.major(),
						unused.second.minor(),
						QString::fromStdString(toStdString(unused.second.code()))
					);
				elements.reconciliations_.emplace_back(
					bankAccount.account_.code_.number().integer(),
					reconciledSurvey.reconciled().balance().major(),
					reconciledSurvey.reconciled().balance().minor(),
					QString::fromStdString(
						toStdString(reconciledSurvey.reconciled().balance().code())
					),
					reconciledSurvey.reconciled().reconciledOn(),
					reconciledSurvey.checksum(),
					move(distributions)
				);
			}
			return elements;
		}()
	};
	qDebug() << "Writing budget elements";
	applicationElement.write(xml);
	const vector<ObserverPtr<const BasicElement>> budgetWriters{
		&wageElements,
		&billElements,
		&debtsElements,
		&goalElements,
		&nontrackElements,
		&accountsElement,
		&ledgersElement,
		&banksElement,
		&reconciliationsElement
	};

	for (const auto& writer: budgetWriters)
		writer->write(xml);
	xml.writeEndElement(); // Closes the application element.
	qDebug() << "Completed writing budget elements";
}
