/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023                  ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#include "ReconciledSurvey.h"
using pecunia::currency::Money;
using ::QString;
using std::map;
using drn::banking::BankAccount;
using drn::banking::presentationText;
using drn::banking::ReconciledBankAccount;
using drn::budgeting::BudgetItemIdentifier;
using drn::foundation::Optional;
using drn::surveying::ReconciledSurvey;
using drn::surveying::ReconciledSurveys;

#include <pecunia/Codes.h>
using pecunia::currency::Iso4217Codes;
#include <pecunia/Math.h>
using pecunia::math::sum;
#include <QCryptographicHash>
using ::QCryptographicHash;
#include <QDate>
using ::QDate;
#include <QObject>
using ::QObject;
#include <QStringBuilder>
// using operator%
#include <Qt>
using Qt::DateFormat;

#include <algorithm>
using std::transform;
#include <iterator>
using std::inserter;
#include <ostream>
using std::ostream;
#include <set>
using std::set;
#include <tuple>
using std::tie;
#include <utility>
using std::move;

#include <accounting/AccountCode.h>
using drn::accounting::AccountCode;
#include <accounting/AccountingErrors.h>
using drn::accounting::AccountError;
#include <accounting/AccountNumber.h>
using drn::accounting::AccountNumber;
#include <banking/BankAccountType.h>
using drn::banking::BankAccountType;
#include <banking/BankName.h>
using drn::banking::BankName;
#include <banking/BankingErrors.h>
using drn::banking::BankError;
#include <budgeting/BudgetingErrors.h>
using drn::budgeting::BudgetItemError;
#include <budgeting/BudgetSource.h>
using drn::budgeting::BudgetSource;
#include <foundation/Error.h>
using drn::foundation::Error;
#include <foundation/PresentationTags.hpp>
using drn::foundation::LocaleAware;
using drn::foundation::WithCurrency;
#include <foundation/PresentationText.h>
using drn::foundation::presentationText;


namespace
{

QString calculateChecksum(
	const BankAccount& ba,
	const Money& balance,
	const Optional<QDate>& reconciledOn,
	const map<BudgetItemIdentifier, Money>& distribution
)
{
	QString value{presentationText(ba, LocaleAware{false})};
	value = value % presentationText(balance, LocaleAware{false}, WithCurrency{true});

	if (reconciledOn.hasValue())
		value = value % reconciledOn->toString(DateFormat::ISODate);

	for (const auto& itemAmount : distribution)
		value = value % presentationText(itemAmount.first)
				% presentationText(itemAmount.second, LocaleAware{false}, WithCurrency{true});
	return QCryptographicHash::hash(value.toUtf8(), QCryptographicHash::Sha256).toBase64();
}

}

//{ ReconciledSurvey

ReconciledSurvey::ReconciledSurvey(BankAccount ba, const Iso4217Codes& currency) :
	ReconciledSurvey{ReconciledBankAccount{ba, currency}, {}, {}}
{}

ReconciledSurvey::ReconciledSurvey(
	ReconciledBankAccount reconciled,
	map<BudgetItemIdentifier, Money> distribution,
	Optional<QString> checksum
) :
	reconciled_{move(reconciled)},
	distribution_{move(distribution)},
	checksum_{move(checksum)}
{
	if (this->checksum_.hasValue())
	{
		const auto calculatedSum{
			calculateChecksum(
				this->reconciled_.bankAccount(),
				this->reconciled_.balance(),
				this->reconciled_.reconciledOn(),
				this->distribution_
			)
		};

		if (this->checksum_ != calculatedSum)
			throw Error{
				QObject::tr(
					"The supplied values for a reconciled account does not match the expected "
						"checksum value, %1."
				).arg(calculatedSum)
			};
	}
}

const ReconciledBankAccount& ReconciledSurvey::reconciled() const noexcept
{
	return this->reconciled_;
}

const map<BudgetItemIdentifier, Money>& ReconciledSurvey::distribution() const noexcept
{
	return this->distribution_;
}

const Optional<QString>& ReconciledSurvey::checksum() const noexcept
{
	return this->checksum_;
}

void ReconciledSurvey::completed(
	const QDate& completedOn,
	const Money& reconciledBalance,
	const map<BudgetItemIdentifier, Money>& distribution
)
{
	const auto distributionBalance{sum(this->reconciled_.balance().code(), distribution)};
	const auto ba{this->reconciled_.bankAccount()};

	if (reconciledBalance != distributionBalance)
		throw BankError{
			ba.bank_,
			QObject::tr("The surveyed balance %1 does not match the unused %2 balance.")
				.arg(presentationText(reconciledBalance))
				.arg(presentationText(distributionBalance))
		};
	this->reconciled_.reconcile(completedOn, reconciledBalance);
	this->distribution_ = move(distribution);
	this->checksum_ = calculateChecksum(
		this->reconciled_.bankAccount(),
		reconciledBalance,
		completedOn,
		distribution
	);
}

bool drn::surveying::operator==(const ReconciledSurvey& lhs, const ReconciledSurvey& rhs)
{
	return tie(lhs.reconciled(), lhs.distribution()) == tie(rhs.reconciled(), rhs.distribution());
}

bool drn::surveying::operator!=(const ReconciledSurvey& lhs, const ReconciledSurvey& rhs)
{
	return ! (lhs == rhs);
}

ostream& drn::surveying::operator<<(ostream& o, const ReconciledSurvey& rs)
{
	o << "Reconciled: (" << rs.reconciled() << "), Distribution: (";
	bool isFirst{true};

	for (const auto& budgetItemAmount : rs.distribution())
		o << (isFirst ? isFirst = false, "" : ", ") << budgetItemAmount.first << "=>"
			<< budgetItemAmount.second;
	return o << ')';
}

void ReconciledSurvey::renamed(const BankName& original, const BankName& newName)
{
	if (this->reconciled_.bankAccount().bank_ != original)
		throw BankError{
			this->reconciled_.bankAccount().bank_,
			QObject::tr("The bank account '%1' is not the surveyed account.")
				.arg(presentationText(original))
		};
	this->reconciled_ = ReconciledBankAccount{
		BankAccount{newName, this->reconciled_.bankAccount().account_},
		this->reconciled_.balance(),
		this->reconciled_.reconciledOn()
	};
	this->checksum_ = calculateChecksum(
		this->reconciled_.bankAccount(),
		this->reconciled_.balance(),
		this->reconciled_.reconciledOn(),
		this->distribution_
	);
}

void ReconciledSurvey::renamed(const AccountNumber& original, const AccountCode& changed)
{
	if (original != this->reconciled_.bankAccount().account_.code_.number())
		throw AccountError{
			this->reconciled_.bankAccount().account_.code_,
			QObject::tr("The account number '%1' is not the surveyed account.")
				.arg(presentationText(original))
		};
	this->reconciled_ = ReconciledBankAccount{
		BankAccount{
			this->reconciled_.bankAccount().bank_,
			BankAccountType{changed, this->reconciled_.bankAccount().account_.type_}
		},
		this->reconciled_.balance(),
		this->reconciled_.reconciledOn()
	};
	this->checksum_ = calculateChecksum(
		this->reconciled_.bankAccount(),
		this->reconciled_.balance(),
		this->reconciled_.reconciledOn(),
		this->distribution_
	);
}

void ReconciledSurvey::renamed(const BudgetItemIdentifier& id, const BudgetSource& newName)
{
	const auto item{this->distribution_.find(id)};

	if (item == this->distribution_.cend())
		throw BudgetItemError{
			newName,
			QObject::tr("The budget item is not a part of the unused balance distribution.")
		};
	const auto result{
		this->distribution_.emplace(BudgetItemIdentifier{item->first.type_, newName}, item->second)
	};
	assert(result.second && "The new item should always be inserted.");
	this->distribution_.erase(item);
	this->checksum_ = calculateChecksum(
		this->reconciled_.bankAccount(),
		this->reconciled_.balance(),
		this->reconciled_.reconciledOn(),
		this->distribution_
	);
}

//}

//{ ReconciledSurveys

ReconciledSurveys::ReconciledSurveys(map<BankAccount, ReconciledSurvey> surveys) :
	map<BankAccount, ReconciledSurvey>{move(surveys)}
{
	for (const auto& accountSurvey : *this)
		if (accountSurvey.first != accountSurvey.second.reconciled().bankAccount())
			throw BankError{
				accountSurvey.first.bank_,
				QObject::tr("The index bank account, %1, must match the surveyed bank account, %2.")
					.arg(presentationText(accountSurvey.first))
					.arg(presentationText(accountSurvey.second.reconciled().bankAccount()))
			};
}

Optional<ReconciledSurvey> ReconciledSurveys::find(const BankAccount& ba) const
{
	return this->find(ba.bank_, ba.account_.code_.number());
}

Optional<ReconciledSurvey> ReconciledSurveys::find(
	const BankName& bn,
	const AccountNumber& an
) const
{
	const auto account{
		find_if(
			this->cbegin(),
				this->cend(),
				[&bn, &an] (const auto& reconciledSurvey)
				{
					return reconciledSurvey.first.bank_ == bn
					&& reconciledSurvey.first.account_.code_ == an;
				}
		)
	};

	if (account == this->cend())
		return {};
	return account->second;
}

bool ReconciledSurveys::upsert(ReconciledSurvey survey)
{
	const auto ba{survey.reconciled().bankAccount()};
	set<BudgetItemIdentifier> idsUsed;
	transform(
		survey.distribution().cbegin(),
		survey.distribution().cend(),
		inserter(idsUsed, idsUsed.begin()),
		[] (const auto& idAmount)
		{
			return idAmount.first;
		}
	);

	for (const auto& accountSurvey : *this)
	{
		if (accountSurvey.first == ba)
			continue;

		for (const auto& idAmount : accountSurvey.second.distribution())
			if ( ! idsUsed.emplace(idAmount.first).second)
				throw BankError{
					ba.bank_,
					QObject::tr(
						"Cannot distribute '%1' into '%2' as it's already distributed in '%3'."
					)
						.arg(presentationText(idAmount.first))
						.arg(presentationText(ba.account_))
						.arg(presentationText(accountSurvey.first))
				};
	}

	if (this->count(ba) == 0)
	{
		const auto emplaced{(*this).emplace(ba, move(survey))};
		assert(emplaced.second && "A new survey should always be inserted.");
		return true;
	}
	else
	{
		(*this)[ba] = move(survey);
		return false;
	}
}

bool drn::surveying::operator==(const ReconciledSurveys& lhs, const ReconciledSurveys& rhs)
{
	if (lhs.size() != rhs.size())
		return false;

	for (const auto& bankAccountReconciledSurvey : lhs)
		if (bankAccountReconciledSurvey.second != rhs.at(bankAccountReconciledSurvey.first))
			return false;
	return true;
}

bool drn::surveying::operator!=(const ReconciledSurveys& lhs, const ReconciledSurveys& rhs)
{
	return ! (lhs == rhs);
}

ostream& drn::surveying::operator<<(ostream& o, const ReconciledSurveys& rss)
{
	o << "Reconciled Surveys (";
	bool isFirst{true};

	for (const auto& bankAccountReconciledSurvey : rss)
		o << (isFirst ? isFirst = false, "" : ", ") << bankAccountReconciledSurvey.first << "=>"
				<< bankAccountReconciledSurvey.second;
	return o << ')';
}

//}
